#include "AbortControllerBinding.h"
#include "AbortSignalBinding.h"
#include "AddonManagerBinding.h"
#include "ChannelWrapperBinding.h"
#include "ChromeUtilsBinding.h"
#include "ConsoleBinding.h"
#include "DOMErrorBinding.h"
#include "DOMExceptionBinding.h"
#include "DOMRequestBinding.h"
#include "DOMStringListBinding.h"
#include "DominatorTreeBinding.h"
#include "ErrorEventBinding.h"
#include "EventBinding.h"
#include "EventTargetBinding.h"
#include "FileReaderBinding.h"
#include "HeapSnapshotBinding.h"
#include "IDBCursorBinding.h"
#include "IDBDatabaseBinding.h"
#include "IDBFactoryBinding.h"
#include "IDBFileHandleBinding.h"
#include "IDBFileRequestBinding.h"
#include "IDBIndexBinding.h"
#include "IDBKeyRangeBinding.h"
#include "IDBMutableFileBinding.h"
#include "IDBObjectStoreBinding.h"
#include "IDBOpenDBRequestBinding.h"
#include "IDBRequestBinding.h"
#include "IDBTransactionBinding.h"
#include "IDBVersionChangeEventBinding.h"
#include "IdleDeadlineBinding.h"
#include "MatchGlobBinding.h"
#include "MatchPatternBinding.h"
#include "MessageChannelBinding.h"
#include "MessageEventBinding.h"
#include "MessagePortBinding.h"
#include "PrecompiledScriptBinding.h"
#include "PromiseDebuggingBinding.h"
#include "StreamFilterBinding.h"
#include "StreamFilterDataEventBinding.h"
#include "StructuredCloneHolderBinding.h"
#include "TCPServerSocketBinding.h"
#include "TCPServerSocketEventBinding.h"
#include "TCPSocketBinding.h"
#include "TCPSocketErrorEventBinding.h"
#include "TCPSocketEventBinding.h"
#include "TextDecoderBinding.h"
#include "TextEncoderBinding.h"
#include "URLSearchParamsBinding.h"
#include "WebExtensionContentScriptBinding.h"
#include "WebExtensionPolicyBinding.h"
#include "WorkerBinding.h"
#include "js/Id.h"
#include "mozilla/dom/WebIDLGlobalNameHash.h"
#include "nsThreadUtils.h"

namespace mozilla {
namespace dom {
static const char *IdString(unsigned int aIndex)
{
  static const char table[] = "AbortController" "\0" "AbortSignal" "\0" "AddonManagerPermissions" "\0" "ChannelWrapper" "\0" "ChromeUtils" "\0" "ChromeWorker" "\0" "ConsoleInstance" "\0" "DOMError" "\0" "DOMException" "\0" "DOMRequest" "\0" "DOMStringList" "\0" "DominatorTree" "\0" "ErrorEvent" "\0" "Event" "\0" "EventTarget" "\0" "FileReader" "\0" "HeapSnapshot" "\0" "IDBCursor" "\0" "IDBCursorWithValue" "\0" "IDBDatabase" "\0" "IDBFactory" "\0" "IDBFileHandle" "\0" "IDBFileRequest" "\0" "IDBIndex" "\0" "IDBKeyRange" "\0" "IDBLocaleAwareKeyRange" "\0" "IDBMutableFile" "\0" "IDBObjectStore" "\0" "IDBOpenDBRequest" "\0" "IDBRequest" "\0" "IDBTransaction" "\0" "IDBVersionChangeEvent" "\0" "IdleDeadline" "\0" "MatchGlob" "\0" "MatchPattern" "\0" "MatchPatternSet" "\0" "MessageChannel" "\0" "MessageEvent" "\0" "MessagePort" "\0" "MozQueryInterface" "\0" "PrecompiledScript" "\0" "PromiseDebugging" "\0" "StreamFilter" "\0" "StreamFilterDataEvent" "\0" "StructuredCloneHolder" "\0" "TCPServerSocket" "\0" "TCPServerSocketEvent" "\0" "TCPSocket" "\0" "TCPSocketErrorEvent" "\0" "TCPSocketEvent" "\0" "TextDecoder" "\0" "TextEncoder" "\0" "URLSearchParams" "\0" "WebExtensionContentScript" "\0" "WebExtensionPolicy" "\0" "Worker" "\0" "console";
  static const uint16_t indices[] = { 0, 16, 28, 52, 67, 79, 92, 108, 117, 130, 141, 155, 169, 180, 186, 198, 209, 222, 232, 251, 263, 274, 288, 303, 312, 324, 347, 362, 377, 394, 405, 420, 442, 455, 465, 478, 494, 509, 522, 534, 552, 570, 587, 600, 622, 644, 660, 681, 691, 711, 726, 738, 750, 766, 792, 811, 818 };
  static_assert(826 <= UINT16_MAX, "string table overflow!");
  return &table[indices[aIndex]];
}

struct SystemProperty
{
  WebIDLGlobalNameHash::ConstructorEnabled enabled;
  ProtoGetter define;
  PinnedStringId id;
};

static SystemProperty properties[] = {
  { nullptr, AbortControllerBinding::GetConstructorObject },
  { nullptr, AbortSignalBinding::GetConstructorObject },
  { AddonManagerPermissionsBinding::ConstructorEnabled, AddonManagerPermissionsBinding::GetConstructorObject },
  { ChannelWrapperBinding::ConstructorEnabled, ChannelWrapperBinding::GetConstructorObject },
  { ChromeUtilsBinding::ConstructorEnabled, ChromeUtilsBinding::GetConstructorObject },
  { ChromeWorkerBinding::ConstructorEnabled, ChromeWorkerBinding::GetConstructorObject },
  { ConsoleInstanceBinding::ConstructorEnabled, ConsoleInstanceBinding::GetConstructorObject },
  { nullptr, DOMErrorBinding::GetConstructorObject },
  { nullptr, DOMExceptionBinding::GetConstructorObject },
  { nullptr, DOMRequestBinding::GetConstructorObject },
  { nullptr, DOMStringListBinding::GetConstructorObject },
  { DominatorTreeBinding::ConstructorEnabled, DominatorTreeBinding::GetConstructorObject },
  { nullptr, ErrorEventBinding::GetConstructorObject },
  { nullptr, EventBinding::GetConstructorObject },
  { nullptr, EventTargetBinding::GetConstructorObject },
  { nullptr, FileReaderBinding::GetConstructorObject },
  { HeapSnapshotBinding::ConstructorEnabled, HeapSnapshotBinding::GetConstructorObject },
  { nullptr, IDBCursorBinding::GetConstructorObject },
  { nullptr, IDBCursorWithValueBinding::GetConstructorObject },
  { nullptr, IDBDatabaseBinding::GetConstructorObject },
  { nullptr, IDBFactoryBinding::GetConstructorObject },
  { nullptr, IDBFileHandleBinding::GetConstructorObject },
  { nullptr, IDBFileRequestBinding::GetConstructorObject },
  { nullptr, IDBIndexBinding::GetConstructorObject },
  { nullptr, IDBKeyRangeBinding::GetConstructorObject },
  { IDBLocaleAwareKeyRangeBinding::ConstructorEnabled, IDBLocaleAwareKeyRangeBinding::GetConstructorObject },
  { nullptr, IDBMutableFileBinding::GetConstructorObject },
  { nullptr, IDBObjectStoreBinding::GetConstructorObject },
  { nullptr, IDBOpenDBRequestBinding::GetConstructorObject },
  { nullptr, IDBRequestBinding::GetConstructorObject },
  { nullptr, IDBTransactionBinding::GetConstructorObject },
  { nullptr, IDBVersionChangeEventBinding::GetConstructorObject },
  { IdleDeadlineBinding::ConstructorEnabled, IdleDeadlineBinding::GetConstructorObject },
  { MatchGlobBinding::ConstructorEnabled, MatchGlobBinding::GetConstructorObject },
  { MatchPatternBinding::ConstructorEnabled, MatchPatternBinding::GetConstructorObject },
  { MatchPatternSetBinding::ConstructorEnabled, MatchPatternSetBinding::GetConstructorObject },
  { nullptr, MessageChannelBinding::GetConstructorObject },
  { nullptr, MessageEventBinding::GetConstructorObject },
  { nullptr, MessagePortBinding::GetConstructorObject },
  { MozQueryInterfaceBinding::ConstructorEnabled, MozQueryInterfaceBinding::GetConstructorObject },
  { PrecompiledScriptBinding::ConstructorEnabled, PrecompiledScriptBinding::GetConstructorObject },
  { PromiseDebuggingBinding::ConstructorEnabled, PromiseDebuggingBinding::GetConstructorObject },
  { StreamFilterBinding::ConstructorEnabled, StreamFilterBinding::GetConstructorObject },
  { StreamFilterDataEventBinding::ConstructorEnabled, StreamFilterDataEventBinding::GetConstructorObject },
  { StructuredCloneHolderBinding::ConstructorEnabled, StructuredCloneHolderBinding::GetConstructorObject },
  { TCPServerSocketBinding::ConstructorEnabled, TCPServerSocketBinding::GetConstructorObject },
  { TCPServerSocketEventBinding::ConstructorEnabled, TCPServerSocketEventBinding::GetConstructorObject },
  { TCPSocketBinding::ConstructorEnabled, TCPSocketBinding::GetConstructorObject },
  { TCPSocketErrorEventBinding::ConstructorEnabled, TCPSocketErrorEventBinding::GetConstructorObject },
  { TCPSocketEventBinding::ConstructorEnabled, TCPSocketEventBinding::GetConstructorObject },
  { nullptr, TextDecoderBinding::GetConstructorObject },
  { nullptr, TextEncoderBinding::GetConstructorObject },
  { nullptr, URLSearchParamsBinding::GetConstructorObject },
  { WebExtensionContentScriptBinding::ConstructorEnabled, WebExtensionContentScriptBinding::GetConstructorObject },
  { WebExtensionPolicyBinding::ConstructorEnabled, WebExtensionPolicyBinding::GetConstructorObject },
  { WorkerBinding::ConstructorEnabled, WorkerBinding::GetConstructorObject },
  { nullptr, consoleBinding::GetConstructorObject },
};

static bool idsInited = false;

bool
SystemBindingInitIds(JSContext* aCx)
{
  MOZ_ASSERT(NS_IsMainThread());

  if (!idsInited) {
    // We can't use range-based for because we need the index to call IdString.
    for (uint32_t i = 0; i < ArrayLength(properties); ++i) {
      if (!properties[i].id.init(aCx, IdString(i))) {
        return false;
      }
    }
    idsInited = true;
  }

  return true;
}

bool
ResolveSystemBinding(JSContext* aCx, JS::Handle<JSObject*> aObj, JS::Handle<jsid> aId, bool* aResolvedp)
{
  MOZ_ASSERT(NS_IsMainThread());
  MOZ_ASSERT(idsInited);

  if (JSID_IS_VOID(aId)) {
    for (const auto& property : properties) {
      if (!property.enabled || property.enabled(aCx, aObj)) {
        if (!property.define(aCx)) {
          return false;
        }
        *aResolvedp = true;
      }
    }
    return true;
  }

  for (const auto& property : properties) {
    if (property.id == aId) {
      if (!property.enabled || property.enabled(aCx, aObj)) {
        if (!property.define(aCx)) {
          return false;
        }
        *aResolvedp = true;
        break;
      }
    }
  }
  return true;
}

bool
MayResolveAsSystemBindingName(jsid aId)
{
  MOZ_ASSERT(NS_IsMainThread());
  MOZ_ASSERT(idsInited);

  for (const auto& property : properties) {
    if (aId == property.id) {
      return true;
    }
  }
  return false;
}

void
GetSystemBindingNames(JSContext* aCx, JS::Handle<JSObject*> aObj, JS::AutoIdVector& aNames, bool aEnumerableOnly, mozilla::ErrorResult& aRv)
{
  MOZ_ASSERT(NS_IsMainThread());

  if (aEnumerableOnly) {
    return;
  }

  if (!SystemBindingInitIds(aCx)) {
    aRv.NoteJSContextException(aCx);
    return;
  }

  for (const auto& property : properties) {
    if (!property.enabled || property.enabled(aCx, aObj)) {
      if (!aNames.append(property.id)) {
        aRv.Throw(NS_ERROR_OUT_OF_MEMORY);
        return;
      }
    }
  }
}

} // namespace dom
} // namespace mozilla

